# coding=utf-8
'''
Created on 23.03.2011

@author: pakula
'''

import os.path
import socket
from uuid import uuid4
import html
import sys
from io import BufferedIOBase, TextIOBase
import io

class TracerError(Exception):pass

class TraceNode:
    def __init__(self, name, children=None, attribs=None):
        self.name = name
        self.attribs = attribs or {}
        self.children = children or []
    
    def append(self, *nodes):
        self.children.extend(nodes)
        
    def dumpAttribs(self):
        tmp = []
        for key, value in self.attribs.items():
            try:
                value = str(value) if value is not None else 'None'
            except:
                value = 'N/A'
            value = html.escape(value, quote=True)
            tmp.extend([' ', key, '="', value if not value is None else '', '"'])
        
        return tmp
    def dumpContents(self, indent):
        result = []
        for kid in self.children:
            if not kid is None: 
                result.extend(kid.dump(None, indent + '  '))
        return result
        
    def dump(self, aspect=None, indent=''):
        tag = [aspect, ":", self.name] if aspect else [self.name]  
        result = [indent, '<']
        result.extend(tag)
        result.extend(self.dumpAttribs())
        contents = self.dumpContents(indent)
        if not contents:
            result.append('/>')
            result.append("\n")
        else:
            result.append('>')
            result.append("\n")
            result.extend(contents)
            result.extend([indent, "</"])
            result.extend(tag)
            result.append(">\n")
        return result
    def string(self, aspect=None, indent=''):
        return ''.join(self.dump(aspect, indent))
        
class TextNode:
    def __init__(self, text):
        self.text = str(text);
    def dump(self, aspect=None, indent=''):
        ''' Замена недопустимой в CDATA последовательности ]]> на 
        ]]]]><![CDATA[>'''
        slices = "]]]]><![CDATA[>".join(self.text.split(']]>'))
        return [indent, '<![CDATA[', slices, ']]>', '\n']
    def string(self, aspect=None, indent=''):
        return ''.join(self.dump(aspect, indent))
    
class Failure(TraceNode):
    def __init__(self, short_desc, detailed_desc=None, traceback=None, depth=None, attribs=None):
        sdnode = TraceNode("short-desc", children=[TextNode(short_desc)])
        ddnode = TraceNode("description", children=[TextNode(detailed_desc)])
        stack = None
        if not traceback is None:
            frames = []
            while traceback:
                frame = traceback.tb_frame
                traceback = traceback.tb_next
                code = frame.f_code
                if 'self' in frame.f_locals:
                    clazz = frame.f_globals['__name__'] + '.' + frame.f_locals['self'].__class__.__name__
                else:
                    clazz = frame.f_globals['__name__']
                frames.append(
                              TraceNode('frame',
                                        attribs={'class' : clazz,
                                                 'method' : code.co_name,
                                                 'source' : code.co_filename,
                                                 'line' : frame.f_lineno}))
            stack = TraceNode('stack', children=frames)

        TraceNode.__init__(self, name="failure", children=[sdnode, ddnode, stack], attribs=attribs)

class SocketStream:
    def __init__(self, socket):
        self.stream = socket
    def __hash__(self):
        return hash(self.stream)
    def close(self):
        if not self.stream is None:
            self.stream.close()
            self.stream = None
    def write(self, string):
        self.stream.send(string.encode())
                
class Tracer:
    __singleton = None
    @staticmethod
    def defaultTracer():
        if Tracer.__singleton is None:
            Tracer.__singleton = Tracer()
        return Tracer.__singleton

    INIT = 0
    RUNNING = 1
    DONE = 2
    def __init__(self):
        self.streams = set()
        self.filters = set()
        self.state = Tracer.INIT
        self.disabledAspects = set()
        self.files = set()
        
    def __del__(self):
        if self.state == self.RUNNING:
            self.stop()
        self.streams.clear()
        

    def addFile(self, filename):
        if self.state != Tracer.INIT:
            raise TracerError("The tracer is already initialized") 
        absolute = os.path.realpath(filename)
        if absolute in self.files:
            return
        file = open(absolute, mode='w', encoding="utf-8")
        self.addStream(file) 
        self.files.add(absolute)
        
    def addSocket(self, ip, port):
        if self.state != Tracer.INIT:
            raise TracerError("The tracer is already initialized") 
        addr = (ip, port)
        if addr in self.sockets: return
        sock = socket.socket()
        try:
            sock.connect(addr)
        except IOError:
            sock.close()
            return
        self.addStream(SocketStream(sock))
        
    def addStream(self, stream):
        self.streams.add(stream)
    
    def enableAspects(self, *aspects):
        self.disabledAspects -= set(aspects)
    def enableAllAspects(self):
        self.disabledAspects.clear()
    def disableAspects(self, *aspects):
        self.disabledAspects |= aspects
    def isEnabledAspect(self, aspect):
        return not aspect in self.disabledAspects
    
    def _write(self, strarg):
        if self.state != self.RUNNING:
            # raise TracerError("The tracer is not running")
            self.start() 
        for stream in self.streams:
            stream.write(strarg)
            
    def start(self):
        if self.state != Tracer.INIT:
            raise TracerError("The tracer is already started")
        self.state = Tracer.RUNNING
        self._write("<?xml version=\"1.0\" encoding=\"utf-8\"?>\n") 
        self._write("<trace format=\"mixed\" version=\"3.0\" t:def-channel=\"0\">\n")
        
    def stop(self):
        if self.state != self.RUNNING:
            raise TracerError("The tracer is not running")
        self._write("</trace>")
        self.state = self.DONE
        for stream in self.streams:
            try:
                stream.close()
            except:
                continue
        self.streams.clear()
        
    def trace(self, node, aspect=None):
        if self.state != Tracer.RUNNING:
            # raise TracerError("The tracer is not running")
            self.start()
        if node is None: return
        self._write(node.string(aspect, '  '))
        
    def _tag(self, action, aspect, tag, parent=None):
        '''Utility method. Traces internal tags tag-start and tag-end'''
        if not self.isEnabledAspect(aspect):
            return
        msg = "  <tag-%s aspect=\"%s\" name=\"%s\"" % (action, aspect, tag)
        if not parent is None:
            msg += " parent=\"%s\"" % parent
        msg += '/>\n'
        self._write(msg)
    def tagStart(self, aspect, tag, parent=None):
        '''Traces start of the given aspect'''
        self._tag('start', aspect, tag, parent)
    def tagEnd(self, aspect, tag, parent=None):
        '''Traces end of the given aspect.''' 
        self._tag('end', aspect, tag, parent)
    
    def traceFailure(self, description, detailed=None, traceback=None, aspect=None, exception=None):
        if self.isEnabledAspect('error'):
            attribs = {}
            if aspect: attribs['aspect'] = aspect
            if hasattr(exception, 'failid'): attribs['failid'] = exception.failid
            self.trace(Failure(description, detailed, traceback=traceback, attribs=attribs), 'error')

    def __enter__(self):
        if self.state is Tracer.INIT:
            self.start()
    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()
        
class DummyTracer(Tracer):
    __singleton = None
    @staticmethod
    def defaultTracer():
        if DummyTracer.__singleton is None:
            DummyTracer.__singleton = DummyTracer()
        return DummyTracer.__singleton

    def __init__(self):
        pass

    def __del__(self):
        pass        

    def addFile(self, filename):
        pass
            
    def addSocket(self, ip, port):
        pass
        
    def addStream(self, stream):
        pass
    
    def enableAspects(self, *aspects):
        pass
    def enableAllAspects(self):
        pass
    def disableAspects(self, *aspects):
        pass
    def isEnabledAspect(self, aspect):
        return True
    
    def start(self):
        pass
        
    def stop(self):
        pass
    
    def trace(self, node, aspect=None):
        pass
        
    def tagStart(self, aspect, tag, parent=None):
        '''Traces start of the given aspect'''
        pass
    
    def tagEnd(self, aspect, tag, parent=None):
        '''Traces end of the given aspect.''' 
        pass
    
    def traceFailure(self, description, detailed=None, traceback=None, aspect=None, excepion=None):
        pass
    def __enter__(self):
        pass
    def __exit__(self, exc_type, exc_value, traceback):
        pass
        
class CoverageTracer:
    def __init__(self, tracer=Tracer.defaultTracer()):
        self.tracer = tracer
    def _processElement(self, el):
        """Служебный метод. Строит узел трассы для элемента покрытия"""
        if isinstance(el, str):
            return TraceNode("element", [], {"name" : el})
        else:
            return TraceNode("element", [self._processElement(e) for e in el[1]], {"name" : el[0]})
    def traceCoverageStructure(self, name, structure, aspect=None):
        '''Трассировка структуры покрытия.
        Структура покрытия задается как список, элементы списка могут быть
        * строка - имя элемента покрытия,
        * последовательность (список или кортеж) - первый элемент последовательности интерпретируется как 
          имя элемента покрытия, строка, второй элемент - как структура подпокрытия.'''
        attribs = { 'name' : name }
        if aspect: attribs['aspect'] = aspect
        node = TraceNode("structure", \
                         [self._processElement(el) for el in structure], \
                         attribs)
        self.tracer.trace(node, 'coverage')
    
    def traceCoverageHit(self, coverageName, elements, aspect=None):
        top = bottom = None
        for el in elements:
            if bottom is None:
                top = bottom = TraceNode('element', [], {'name' : el})
            else:
                tmp, bottom = bottom, TraceNode('element', [], {'name' : el})
                tmp.children = [bottom]
        top.attribs['cs'] = coverageName
        if not aspect is None:
            top.attribs['aspect'] = aspect
        self.tracer.trace(top, 'coverage')
    
class TransitionScope:
    def __init__(self, fsmtracer, name):   
        self.name = name
        self.fsmtracer = fsmtracer
    def __enter__(self):
        self.fsmtracer.transitionStart(self.name)
    def __exit__(self, exc_type, exc, traceback):
        self.fsmtracer.transitionEnd(self.name)
        
class FSMTracer:
    COUNT = 0
    @staticmethod
    def nextID():
        FSMTracer.COUNT += 1
        id = FSMTracer.COUNT
        return "dfsm.%d" % id

    def __init__(self, scenarioName, tracer=Tracer.defaultTracer(), id=None):
        self.id = id or FSMTracer.nextID()
        self.tracer = tracer
        self.scenarioName = scenarioName
        self.aspect = 'fsm'
        
    def isAllowed(self):
        return self.tracer.isEnabledAspect(self.aspect)
    
    def trace(self, tag, children=[], attribs={}):
        self.tracer.trace(
                          TraceNode(tag, children, attribs),
                          self.aspect
                          )
    def start(self):
        if not self.isAllowed(): return
        self.tracer.tagStart(self.aspect, self.id)
        self.trace('start', attribs={"name" : self.id, "full-name" : self.scenarioName})
    
    def state(self, stateId, stateText=None):
        if not self.isAllowed():
            return;
        if stateText is None:
            stateText = stateId
        self.trace("state", attribs={"id" : stateId, "name" : stateText, "fsm" : self.id})

    def transitionStart(self, scenFuncName):
        if not self.isAllowed():
            return
        self.tracer.tagStart(self.aspect, scenFuncName, self.id)
        self.trace("start-transition", attribs={ "name" : scenFuncName, "fsm" : self.id})
        
    def transitionEnd(self, scenFuncName):
        if not self.isAllowed():
            return
        self.trace("end-transition", attribs={ "name" : scenFuncName, "fsm" : self.id})
        self.tracer.tagEnd(self.aspect, scenFuncName, self.id)
        
    def transition(self, scenFuncName):
        '''Возвращает контекстный объект для использования в инструкции with.
        При входе в with вызывается метод transitionStart, а при выходе - transitionEnd.'''
        return TransitionScope(self, scenFuncName)
        
    def iterationVariable(self, name, value):
        if not self.isAllowed():
            return
        self.trace("iteration-variable",
                   attribs={ "value" : value, "fsm" : self.id,
                   "type" : type(value).__name__, "name" : name }
                   )
    
    def summary(self, summary):
        children = []
        for (key, value) in summary.items():
            children.append(TraceNode('property', attribs={'name' : key, 'value' : value}))
        self.trace("summary", children, { "fsm" : self.id })
    
    def browserMessage(self, level, message):
        self.trace("browser-message", [TextNode(message)], { "fsm" : self.id, 'level' : level })
    
    def end(self):
        self.trace('end', attribs={"name" : self.id})
        self.tracer.tagEnd(self.aspect, self.id)

    def __enter__(self):
        self.start()
    def __exit__(self, exc_type, exc, traceback):
        self.end()
        
class _RequirementContext:
    @staticmethod
    def _failid():
        return uuid4()
     
    def __init__(self, name_slots, rtracer=None):
        self.rtracer = rtracer
        self.reqname = name_slots
    def __enter__(self):
        pass
    def __exit__(self, exc_type, exc, traceback):
        if exc is None:
            self.rtracer.ok(*self.reqname)
        else:
            if not hasattr(exc, 'failid'):
                exc.failid = self._failid()
            message = exc.args
            self.rtracer.fail(message, *self.reqname, failid=exc.failid)
            
class RequirementTracer:
    def __init__(self, tracer=Tracer.defaultTracer(), defaultDB=None):
        '''Создает трассировщик требований.
        База требований по-умолчанию задается в виде (id,name,location), аналогично одноменным параметрам метода db.
        '''
        self.tracer = tracer
        self.defaultDB = defaultDB
        
    def db(self, dbid=None, name=None, location=None):
        if dbid is None:
            if self.defaultDB is None:
                raise Exception('Identified of the requirement DB is not defined')
            dbid = self.defaultDB[0]
            
        if name is None:
            try:
                name = self.defaultDB[1]
            except:
                name = dbid

        if location is None:
            try:
                location = self.defaultDB[2]
            except:
                pass
        
        attribs = {'dbid' :dbid, 'name' : name }
        if not location is None:
            attribs['location'] = location
        self.tracer.trace(TraceNode('db', attribs=attribs), 'requirement')
        
    def ok(self, *name_slots):
        assert len(name_slots) > 0
        
        name = '/'.join(name_slots)
        attribs = { 'name' : name, 'db': self.defaultDB[0] }
        self.tracer.trace(TraceNode('pass', attribs=attribs), 'requirement')
        
    def fail(self, message, *name_slots, **kwargs):
        failid = kwargs.get('failid', None)
        assert len(name_slots) > 0
        
        if message is None:
            message = ''
        name = '/'.join(name_slots)
        attribs = { 'name' : name, 'db': self.defaultDB[0] }
        if failid:
            attribs['failid'] = failid 
        self.tracer.trace(TraceNode('fail', attribs=attribs, children=[TextNode(str(message))]),
                          'requirement')
        
    def requirement(self, *name_slots):
        return _RequirementContext(name_slots, self)

class TestTracer:
    def __init__(self, name, tracer=Tracer.defaultTracer()):
        self.name = name
        self.tracer = tracer
        self.verdict = True
    def start(self):
        self.tracer.trace(
                          TraceNode('start', attribs={'name':self.name}),
                          'test'
                          )
    def end(self, verdict):
        self.tracer.trace(
                          TraceNode('end', attribs={'name':self.name, 'verdict' : verdict}),
                          'test'
                          )
    def setVerdict(self, verdict, override=False):
        'Значение вердикта изменяется только с True на False, или при установленном параметре override'
        if override or self.verdict:
            self.verdict = verdict
    def __enter__(self):
        self.start()
    def __exit__(self, exc_type, exc, traceback):
        self.end(self.verdict)

class LogTracer:
    def __init__(self, tracer=Tracer.defaultTracer(), delimiter=' '):
        self._tracer = tracer
        self._delimiter = delimiter
    
    def log(self, *args):
        text = self._delimiter.join([str(x) for x in args])
        textNode = TextNode(text)
        traceNode = TraceNode('log', [textNode])
        self._tracer.trace(traceNode)

class StdStreamInterceptor(TextIOBase):
    newlines = getattr(sys.stdout, "newlines", None)
    
    def __init__(self, tracer=Tracer.defaultTracer()):
        self.tracer = LogTracer(tracer)
        self.lastLine = None
        
    def close(self, *args, **kwargs):
        self.closed = True
    
    def _not_implemented(self,*args,**kwargs):
        raise io.UnsupportedOperation
    
    def detach(self, *args, **kwargs):
        self.close()
        return None

    def fileno(self):
        raise NotImplementedError
    
    def flush(self, *args, **kwargs):
        if self.lastLine is not None:
            self.tracer.log('flush')
            self.tracer.log(self.lastLine)
            self.lastLine = None
        
    def isatty(self, *args, **kwargs):
        return False
    
    def readable(self, *args, **kwargs):
        return False
    
    read = _not_implemented
    read1 = _not_implemented
    readinto = _not_implemented
    readline = _not_implemented
    readlines = _not_implemented
        
    seek = _not_implemented
    def seekable(self, *args, **kwargs):
        return False
    
    def truncate(self, *args, **kwargs):
        pass
    
    def writable(self, *args, **kwargs):
        return True
    
    def write(self, s):
        if self.lastLine is None: 
            self.lastLine = ''
        self.lastLine += s
        if self.newlines is None or not isinstance(self.newlines, str):
            self.flush()
        else:
            if self.newlines in self.lastLine:
                lines = self.lastLine.split(self.newlines)
                self.lastLine, lines = lines[-1], lines[:-1]
                self.tracer.log(*lines)
    def writelines(self,*strs):
        self.tracer.log(*strs)
class NonclosableStream:
    'Utility class for testing - it wraps existing stream and prevents it from closing when tracer is done.'
    def __init__(self, stream):
        self.stream = stream
    def __hash__(self):
        return hash(self.stream)
    def close(self): pass
    
    def write(self, strarg):
        self.stream.write(strarg)

def defaultTracer():
    return Tracer.defaultTracer()

tracer = Tracer.defaultTracer()

if __name__ == '__main__':
    tracer = defaultTracer()
    tracer.addStream(NonclosableStream(sys.stderr))
    sys.stdout = StdStreamInterceptor()
    print("Test of the lines")